// camera class
// -Kyoung Shin Park (2008 Spring)
//////////////////////////////////////////////////////////////////////
#include <math.h>
#include <stdio.h>
#include "camera.h"

namespace oglclass
{

camera::camera(double fovy, double aspect, double znear, double zfar, cameraType type)
{
	fovy_ = fovy;
    aspect_ = aspect;
    near_ = znear;
    far_ = zfar;

	cameraType_ = type;

	pos_   = vector3(0.0f, 0.0f, 0.0f);
	right_ = vector3(1.0f, 0.0f, 0.0f);
	up_    = vector3(0.0f, 1.0f, 0.0f);
	look_  = vector3(0.0f, 0.0f, 1.0f);
}

camera::~camera()
{
}

void camera::setCameraType(cameraType type)
{
	cameraType_ = type;
}

void camera::setPosition(float x, float y, float z)
{
	pos_.vec[0] = x; pos_.vec[1] = y; pos_.vec[2] = z;
}

void camera::setPosition(const vector3& pos)
{
	pos_ = pos;
}

void camera::setTarget(float x, float y, float z)
{
	look_.vec[0] = pos_.vec[0] - x; 
	look_.vec[1] = pos_.vec[1] - y; 
	look_.vec[2] = pos_.vec[2] - z;
	printf("look_ %f %f %f\n", look_.vec[0], look_.vec[1], look_.vec[2]);
}

void camera::setTarget(const vector3& pos)
{
	look_ = pos_ - pos;
	printf("look_ %f %f %f\n", look_.vec[0], look_.vec[1], look_.vec[2]);
}

void camera::setRight(float x, float y, float z)
{
	right_.vec[0] = x; right_.vec[1] = y; right_.vec[2] = z;
}

void camera::setRight(const vector3& vec)
{
	right_ = vec;
}

void camera::setUp(float x, float y, float z)
{
	up_.vec[0] = x; up_.vec[1] = y; up_.vec[2] = z;
}

void camera::setUp(const vector3& vec)
{
	up_ = vec;
}

void camera::setLook(float x, float y, float z)
{
	look_.vec[0] = x; look_.vec[1] = y; look_.vec[2] = z;
}

void camera::setLook(const vector3& vec)
{
	look_ = vec;
}

vector3 camera::getPosition(void)
{
	return pos_;
}

vector3 camera::getRight(void)
{
	return right_;
}

vector3 camera::getUp(void)
{
	return up_;
}

vector3 camera::getLook(void)
{
	return look_;
}

void camera::walk(float units)
{
	// move only on xz plane for land object
	if( cameraType_ == DRIVE )
		pos_ += vector3(look_.vec[0], 0.0f, look_.vec[2]) * units;

	if( cameraType_ == FLY )
		pos_ += look_ * units;
}

void camera::strafe(float units)
{
	// move only on xz plane for land object
	if( cameraType_ == DRIVE )
		pos_ += vector3(right_.vec[0], 0.0f, right_.vec[2]) * units;

	if( cameraType_ == FLY )
		pos_ += right_ * units;
}

void camera::fly(float units)
{
	// move only on y-axis for land object
	if( cameraType_ == DRIVE )
		pos_.vec[1] += units;

	if( cameraType_ == FLY )
		pos_ += up_ * units;
}

void camera::pitch(float angle)
{
	matrix4x4 T;
	T.rotate(angle, right_);

	// rotate up_ and look_ around right_ vector
	T.transformVector(&up_);
	T.transformVector(&look_);
}

void camera::yaw(float angle)
{
	matrix4x4 T;

	// rotate around world y (0, 1, 0) always for land object
	if( cameraType_ == DRIVE )
		T.rotate(angle, 'y');

	// rotate around own up vector for aircraft
	if( cameraType_ == FLY )
		T.rotate(angle, up_);

	// rotate _right and _look around _up or y-axis
	T.transformVector(&right_);
	T.transformVector(&look_);
}

void camera::roll(float angle)
{
	// only roll for aircraft type
	if( cameraType_ == FLY )
	{
		matrix4x4 T;
		T.rotate(angle, look_);

		// rotate _up and _right around _look vector
		T.transformVector(&right_);
		T.transformVector(&up_);
	}
}

void camera::setAspect(double aspect)
{
    aspect_ = aspect;
}

void camera::lookAt(vector3& pos, vector3& target, vector3& up)
{
	vector3 L = pos - target;
	L.normalize();

	vector3 R;
	R = vector3::crossProduct(up, L);
	R.normalize();

	vector3 U;
	U = vector3::crossProduct(L, R);
	U.normalize();

	//printf("R %f %f %f\n", R[0], R[1], R[2]);
	//printf("U %f %f %f\n", U[0], U[1], U[2]);
	//printf("L %f %f %f\n", L[0], L[1], L[2]);

	pos_   = pos;
	right_ = R;
	up_    = U;
	look_  = L;
}

void camera::applyProjection(void)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(fovy_, aspect_, near_, far_);
    glMatrixMode(GL_MODELVIEW);
}

void camera::apply(void) 
{
    float V[16];

	applyProjection();

	glLoadIdentity();

	// Keep camera's axes orthogonal to eachother
	look_.normalize();

	up_ = vector3::crossProduct(look_, right_);
	up_.normalize();

	right_ = vector3::crossProduct(up_, look_);
	right_.normalize();
	
	//printf("\n");
	//printf("right_ %f %f %f\n", right_[0], right_[1], right_[2]);
	//printf("up_ %f %f %f\n", up_[0], up_[1], up_[2]);
	//printf("look_ %f %f %f\n", look_[0], look_[1], look_[2]);
	//printf("\n");

	// Build the view matrix:
	float x = -vector3::dotProduct(right_, pos_);
	float y = -vector3::dotProduct(up_, pos_);
	float z = -vector3::dotProduct(look_, pos_);

	V[0] = right_.vec[0]; V[1] = up_.vec[0]; V[2] = look_.vec[0];  V[3] = 0.0f;
	V[4] = right_.vec[1]; V[5] = up_.vec[1]; V[6] = look_.vec[1];  V[7] = 0.0f;
	V[8] = right_.vec[2]; V[9] = up_.vec[2]; V[10] = look_.vec[2]; V[11] = 0.0f;
	V[12] = x;            V[13] = y;         V[14] = z;            V[15] = 1.0f;

	//printf("%f %f %f\n", V[0], V[1], V[2]);
	//printf("%f %f %f\n", V[4], V[5], V[6]);
	//printf("%f %f %f\n", V[8], V[9], V[10]);
	//printf("%f %f %f\n", V[12], V[13], V[14]);

	glMultMatrixf(V);
}

} // namespace